# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from StringIO import StringIO

from cubicweb.web import Redirect
from cubicweb.devtools.testlib import CubicWebTC


class ViewsTC(CubicWebTC):

    def test_flat_scheme_concepts_import(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'musique')
            cnx.commit()
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            # simply test the form properly render and is well formed
            self.view('skos.scheme.import', rset=scheme.as_rset(), req=req, template=None)
            req.form = self.fake_form('skos.scheme.import', {
                'stream': ('filename.txt', StringIO('\n\nélectro\nhip-hop\nrap\njazz\nclassique\n')),
                'encoding': u'utf-8',
                'language_code': u'fr',
                'format': u'simple',
                'delimiter': u'tab',
                }, [(scheme, {})])
            # now actually tests the import, using scheme.view and not self.view which doesn't like
            # exception, even Redirect
            self.assertRaises(Redirect, scheme.view, 'skos.scheme.import')
            self.assertEqual(set(c.dc_title() for c in scheme.top_concepts),
                             set(u'électro hip-hop rap jazz classique'.split()))
            self.assertEqual(set(l.language_code for c in scheme.top_concepts for l in c.pref_label),
                             set(['fr']))

    def test_lcsv_scheme_concepts_import(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'gni')
            cnx.commit()
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            # simply test the form properly render and is well formed
            self.view('skos.scheme.import', rset=scheme.as_rset(), req=req,
                      template=None)
            fname = 'lcsv_example_shortened.csv'
            posted = {'stream': (fname, open(self.datapath(fname))),
                      'encoding': u'utf-8',
                      'language_code': u'fr',
                      'delimiter': u'tab',
                      'format': u'lcsv'}
            req.form = self.fake_form('skos.scheme.import', posted, [(scheme, {})])
            # now actually tests the import, using scheme.view and not self.view which doesn't like
            # exception, even Redirect
            self.assertRaises(Redirect, scheme.view, 'skos.scheme.import')
        # check that the concept were added
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.find('ConceptScheme', title=u'gni').one()
            self.assertEqual(len(scheme.top_concepts), 2)
            concepts = cnx.find('Concept')
            self.assertEqual(len(concepts), 5)
            concept1 = cnx.find('Concept',
                                definition="Définition de l'organisation politique de l'organisme,"
                                ).one()
            label = concept1.pref_label[0]
            self.assertEqual(len(concept1.pref_label), 1)
            self.assertEqual((label.label,label.language_code),('Vie politique', 'fr'))
            self.assertEqual(len(concept1.narrower_concept), 2)
            concept2 = cnx.find('Concept',
                                definition="Création volontaire ou en application de la loi").one()
            self.assertEqual(concept2.broader_concept[0], concept1)



if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



"""cubicweb-skos application package

"SKOS implementation for cubicweb"
"""

def to_unicode(obj):
    """Turn some object (usually an exception) to unicode"""
    try:
        # The exception message might already be a unicode string.
        return unicode(obj)
    except UnicodeDecodeError:
        return str(obj).decode(errors='ignore')

class ExtEntity(object):
    """Transitional representation of an entity for use in data importer"""

    def __init__(self, etype, extid, values=None):
        self.etype = etype
        self.extid = extid
        if values is None:
            values = {}
        self.values = values
        self._schema = None

    def __repr__(self):
        return '<%s %s %s>' % (self.etype, self.extid, self.values)

    def prepare(self, schema):
        """Prepare an external entity for later insertion:

        * ensure attributes and inlined relations have a single value
        * turn set([value]) into value and remove key associated to empty set
        * remove non inlined relations and return them as a [(e1key, relation, e2key)] list

        Return a list of non inlined relations that may be inserted later, each relations defined by
        a 3-tuple (subject extid, relation type, object extid).

        Take care the importer may call this method several times.
        """
        if self._schema is not None: # already prepared
            assert self._schema is schema
            return ()
        self._schema = schema
        eschema = schema.eschema(self.etype)
        deferred = []
        entity_dict = self.values
        for rtype in list(entity_dict):
            rschema = schema.rschema(rtype)
            if rschema.final or rschema.inlined:
                assert len(entity_dict[rtype]) <= 1, \
                    "more than one value for attribute %s: %s (%s)" % (rtype, entity_dict[rtype],
                                                                       self.extid)
                if entity_dict[rtype]:
                    entity_dict[rtype] = entity_dict[rtype].pop()
                    if (rschema.final and eschema.has_metadata(rtype, 'format')
                            and not rtype + '_format' in entity_dict):
                        entity_dict[rtype + '_format'] = u'text/plain'
                else:
                    del entity_dict[rtype]
            else:
                for target_extid in entity_dict[rtype]:
                    deferred.append((self.extid, rtype, target_extid))
                del entity_dict[rtype]
        return deferred

    def is_ready(self, extid2eid):
        """Return True if the ext entity is ready, i.e. has all the URIs used in inlined relations
        currently existing.
        """
        entity_dict = self.values
        assert self._schema, 'prepare() method should be called first'
        schema = self._schema
        for rtype in entity_dict:
            rschema = schema.rschema(rtype)
            if not rschema.final:
                # _prepare_extentity should drop other cases from the entity dict
                assert rschema.inlined
                if not entity_dict[rtype] in extid2eid:
                    return False
        # entity is ready, replace all relation's extid by eids
        for rtype in entity_dict:
            rschema = schema.rschema(rtype)
            if rschema.inlined:
                entity_dict[rtype] = extid2eid[entity_dict[rtype]]
        return True



# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from StringIO import StringIO

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.entities import CSVIndentationError, CSVDecodeError


def get_narrower_concepts(concept):
    """Get a dictionnary matching the name of the concept to the dictionnary of its narrower
    concepts (recursive)
    """
    narrower_concepts = {}
    for concept in concept.narrower_concept:
        narrower_concepts[concept.dc_title()] = get_narrower_concepts(concept)
    return narrower_concepts

def build_result_dict(scheme):
    """Create a hierarchy with the concept names in a dictionary"""
    concepts = scheme.top_concepts
    result = {}
    for concept in concepts:
        result[concept.dc_title()] = get_narrower_concepts(concept)
    return result


class ConceptSchemeTC(CubicWebTC):

    def test_top_concepts(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'some classification')
            c1 = cnx.create_entity('Concept', in_scheme=scheme)
            cnx.create_entity('Label', label=u'hip', language_code=u'fr',
                              pref_label_of=c1)
            c2 = cnx.create_entity('Concept', in_scheme=scheme, broader_concept=c1)
            cnx.create_entity('Label', label=u'hop', language_code=u'fr',
                              pref_label_of=c2)
            cnx.commit()
            self.assertEqual(set(x.eid for x in scheme.top_concepts),
                             set((c1.eid,)))

    def test_add_concepts_from_file_interdoc(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'interdoc')
            cnx.commit()
        csv_file = self.datapath('thesaurus_interdoc_7_hierageneTab_shortened.csv')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            with open(csv_file) as sourcefile:
                scheme.add_concepts_from_file(sourcefile, u'utf-8', u'fr', u',')
            result = build_result_dict(scheme)
        expected = {
            u'ADMINISTRATION': {
                u'ACTION PUBLIQUE': {
                    u'ACTE ADMINISTRATIF': {
                        u'ACTE CREATEUR DE DROITS': {},
                        u'ACTE UNILATERAL': {
                            u'VISA': {},
                        },
                        u'CONTRAT PUBLIC': {
                            u'CONTRAT ADMINISTRATIF': {
                                u'CLAUSE EXORBITANTE': {},
                                u'COTRAITANCE': {
                                    u'GROUPEMENT D\'ENTREPRISES': {
                                        u'GROUPEMENT CONJOINT': {},
                                        u'GROUPEMENT SOLIDAIRE': {},
                                    },
                                },
                                u'EQUILIBRE FINANCIER': {
                                    u'FAIT DU PRINCE': {}
                                },
                                u'VENTE EN L\'ETAT FUTUR D\'ACHEVEMENT': {},
                            },
                            u'CONVENTION D\'EXPLOITATION': {},
                            u'PARTENARIAT PUBLIC-PRIVE': {
                                u'SOCIETE LOCALE DE PARTENARIAT': {},
                            },
                        },
                    },
                },
            },
            u'AMENAGEMENT': {
                u'AMENAGEMENT DU TERRITOIRE': {
                    u'AMENAGEMENT DE LA MONTAGNE': {
                        u'PROTECTION DE LA MONTAGNE': {
                            u'RTM': {},
                        },
                        u'UTN': {},
                        u'ZONE MONTAGNE': {},
                    },
                    u'AMENAGEMENT DU LITTORAL': {
                        u'BANDE LITTORALE DES 100 METRES': {},
                    },
                },
            },
        }
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_ok(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        csv_file = self.datapath('hierarchical_csv_example_shortened.csv')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            with open(csv_file) as sourcefile:
                scheme.add_concepts_from_file(sourcefile, u'utf-8', u'fr', u'\t')
            result = build_result_dict(scheme)
        expected = {u'APPELLATION DES LOIS ET DES RAPPORTS':
                    {u'LOIS RAPPORTS JURISPRUDENCE':
                     {u'APPELLATION DE DECISIONS DE JURISPRUDENCE':
                      {u'ARRET BERKANI': {}, u'ARRET TERNON': {}},
                      u'APPELLATION DES RAPPORTS':
                      {u'RAPPORT ARTHUIS': {}, u'LOLF': {}, u'LOV': {}}
                     }
                    },
                    u'LISTE DES MOTS OUTILS':
                    {u'CHAPITRE MOTS OUTILS':
                     {u'MOTS OUTILS':
                      {u'ABATTEMENT': {}, u'ACCORD': {}}
                     }
                    }
                   }
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_sep_inside_concept(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n\tti\tti\n\ttata')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u'\t')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'ti\tti': {},
                              u'tata' : {}}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_decode_error(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        with open(self.datapath('bad_encoding.csv')) as fobj:
            with self.admin_access.web_request() as req:
                scheme = req.entity_from_eid(scheme.eid)
                with self.assertRaises(CSVDecodeError) as cm:
                    scheme.add_concepts_from_file(fobj, u'utf-8', u'fr', u',')
                self.assertEqual(cm.exception.line, 3)
                # Now with the good encoding.
                fobj.seek(0)
                scheme.add_concepts_from_file(fobj, u'latin1', u'fr', u',')
                result = build_result_dict(scheme)
        expected = {u'voici mes concepts en latin1': {u'celui-ci est bon': {},
                                                      u'celui-là n\'est pas bon': {}}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_wrong_indentation(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n\ttiti\n\t\t\ttata')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            with self.assertRaises(CSVIndentationError) as cm:
                scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u'\t')
            self.assertEqual(cm.exception.line, 3)

    def test_add_concepts_from_file_multiple_deindentation(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n\ttiti\n\t\ttata\ntutu')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u'\t')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'titi': {u'tata' : {}}},
                    u'tutu': {}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_sep_comma(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n,titi\n,,tata\n,,  tati  \n tutu \n,tuti\n,,titu\n,toti\n')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u',')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'titi': {u'tata' : {}, u'tati': {}}},
                    u'tutu': {u'tuti': {u'titu': {}}, u'toti': {}}}
        self.assertEqual(result, expected)

    def test_add_concepts_from_file_sep_space(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'rapport')
            cnx.commit()
        rapport = StringIO('toto\n titi   iti\n  tata\n  tati  \ntutu \n tuti\n  titu\n toti\n')
        with self.admin_access.web_request() as req:
            scheme = req.entity_from_eid(scheme.eid)
            scheme.add_concepts_from_file(rapport, u'utf-8', u'fr', u' ')
            result = build_result_dict(scheme)
        expected = {u'toto': {u'titi   iti': {u'tata' : {}, u'tati': {}}},
                    u'tutu': {u'tuti': {u'titu': {}}, u'toti': {}}}
        self.assertEqual(result, expected)


class ConceptTC(CubicWebTC):
    def setUp(self):
        super(ConceptTC, self).setUp()
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'some classification')
            c1 = cnx.create_entity('Concept', in_scheme=scheme)
            cnx.create_entity('Label', label=u'hîp', language_code=u'fr-CA',
                              pref_label_of=c1)
            cnx.create_entity('Label', label=u'hip', language_code=u'en',
                              pref_label_of=c1)
            cnx.create_entity('Label', label=u'hop', language_code=u'fr',
                              alt_label_of=c1)
            cnx.commit()
        self.c1_eid = c1.eid

    def test_dc_title(self):
        with self.admin_access.client_cnx() as cnx:
            c1 = cnx.entity_from_eid(self.c1_eid)
            self.assertEqual(c1.dc_title(), u'hip')
            cnx.lang = 'fr'
            self.assertEqual(c1.dc_title(), u'hîp')

    def test_labels(self):
        with self.admin_access.client_cnx() as cnx:
            c1 = cnx.entity_from_eid(self.c1_eid)
            self.assertEqual(c1.labels,
                             {u'en': u'hip', u'fr': u'hîp'})


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-skos entity classes"""

from logilab.common.decorators import cachedproperty

from cubicweb.predicates import is_instance
from cubicweb.entities import AnyEntity, fetch_config
from cubicweb.entities.adapters import ITreeAdapter


def _add_concept(scheme, label, language_code, **kwargs):
    cnx = scheme._cw
    concept = cnx.create_entity('Concept', in_scheme=scheme, **kwargs)
    cnx.create_entity('Label', label=label, language_code=language_code,
                      pref_label_of=concept)
    # hooks do use existing labels, clear cache here to ensure the label is considered
    concept.cw_clear_all_caches()
    return concept

class CSVParseError(Exception):
    """Error during CSV parsing, with a line information."""

    def __init__(self, line):
        super(CSVParseError, self).__init__()
        self.line = line


class CSVIndentationError(CSVParseError):
    """Invalid indentation (level was reduced by more than one indentation)."""


class CSVDecodeError(CSVParseError):
    """Decode exception, probably due to a wrongly specified encoding."""


class ConceptScheme(AnyEntity):
    __regid__ = 'ConceptScheme'
    fetch_attrs, cw_fetch_order = fetch_config(('title', 'cwuri'))

    def dc_title(self):
        if self.title:
            return self.title
        return self.cwuri

    @cachedproperty
    def top_concepts_rset(self):
        return self._cw.execute(
            'Any C,CU WHERE C cwuri CU, C in_scheme X, NOT C broader_concept SC, X eid %(x)s',
            {'x': self.eid})

    @property
    def top_concepts(self):
        return list(self.top_concepts_rset.entities())

    def add_concept(self, label, language_code=u'en', **kwargs):
        """Add a top-concept to this scheme"""
        return _add_concept(self, label, language_code, **kwargs)

    def add_concepts_from_file(self, stream, encoding, language_code, delimiter):
        """Read a stream and create the listed concepts inside the ConceptScheme

        'delimiters' are considered as hierarchical information. There must be a concept per line,
        each line starting by N delimiters (indicating the hierarchical level) or nothing if the
        concept has no parent.

        Example (delimiter = u';')
        titi
        ;toto
        ;;tata
        --> ok: titi is a top-concept of the scheme, toto is a narrower concept of titi, tata is a 
        narrower concept of toto

        titi
         ;toto
        --> 'titi' and ' ;toto' will be considered as two concepts of the scheme.
        """
        # This ordered list behaves like a state machine. It will contain all the concepts from
        # the conceptscheme to the more recent broader concept. The last element is consequently
        # the current parent concept. When de-indenting, the concepts are popped.
        level = -1
        broaderconcepts = [(self, level)]
        for nline, line in enumerate(stream, 1):
            try:
                line = line.rstrip().decode(encoding)
            except UnicodeDecodeError:
                raise CSVDecodeError(nline)
            if not line:
                continue
            # Find the label by removing leading delimiters (and spaces), then
            # remove trailing delimiters and spaces.
            label = line.lstrip(delimiter + ' ').rstrip(delimiter + ' ')
            if not label:
                # The line is full of delimiter.
                continue
            next_level = line[:line.index(label)].count(delimiter)
            if next_level - level > 1:
                # Concept must be at much one level below its parent.
                raise CSVIndentationError(nline)
            elif next_level <= level:
                # Walk back levels.
                while next_level <= level:
                    _, level = broaderconcepts.pop()
            concept = broaderconcepts[-1][0].add_concept(label, language_code)
            broaderconcepts.append((concept, level))
            level = next_level


class Concept(AnyEntity):
    __regid__ = 'Concept'
    fetch_attrs, cw_fetch_order = fetch_config(('cwuri',))

    def dc_title(self):
        return self.label()

    @property
    def parent_concept(self):
        return self.broader_concept[0] if self.broader_concept else None

    @property
    def scheme(self):
        return self.in_scheme[0]

    @property
    def labels(self):
        return dict((l.short_language_code, l.label) for l in self.pref_label)

    def label(self, language_code=None, default_language_code='en'):
        if language_code is None:
            language_code = self._cw.lang
        try:
            return self.labels[language_code[:2].lower()]
        except KeyError:
            try:
                return self.labels[default_language_code[:2].lower()]
            except KeyError:
                # pick one random label
                return self.labels.values()[0]

    def add_concept(self, label, language_code=u'en'):
        """Add a sub-concept to this concept"""
        return _add_concept(self.scheme, label, language_code, broader_concept=self)


class Label(AnyEntity):
    __regid__ = 'Label'
    fetch_attrs, cw_fetch_order = fetch_config(('language_code', 'label'))

    @property
    def short_language_code(self):
        """Return the 2 letters language code for this label"""
        if self.language_code is None:
            return None
        return self.language_code[:2].lower()


class ConceptITreeAdapter(ITreeAdapter):
    """ITree adapater for Concept"""
    __select__ = is_instance('Concept')
    tree_relation = 'broader_concept'


from yams import ValidationError
from cubicweb import UniqueTogetherError
from cubicweb.entities import adapters
from cubicweb.predicates import ExpectedValuePredicate

_ = unicode

# monkeypatched from cubicweb waiting for #5100373 to be integrated
class IUserFriendlyUniqueTogether(adapters.IUserFriendlyUniqueTogether):

    def raise_user_exception(self):
        rtypes = self.exc.rtypes
        errors = {}
        msgargs = {}
        i18nvalues = []
        for rtype in rtypes:
            errors[rtype] = _('%(KEY-rtype)s is part of violated unicity constraint')
            msgargs[rtype + '-rtype'] = rtype
            i18nvalues.append(rtype + '-rtype')
        errors[''] = _('some relations violate a unicity constraint')
        raise ValidationError(self.entity.eid, errors, msgargs=msgargs, i18nvalues=i18nvalues)


class unique_together_error_for(ExpectedValuePredicate):
    """Return 1 if exception given as `exc` in the input context is an UniqueTogetherError implying one
    of the relation type given on instanciation of this predicate.
    """
    def __call__(self, cls, req, exc=None, **kwargs):
        if exc is not None and isinstance(exc, UniqueTogetherError):
            return len(frozenset(exc.rtypes) & self.expected)
        return 0


class IUserFriendlyUniqueTogetherPrefLabelLanguageCode(IUserFriendlyUniqueTogether):
    __select__ =  (IUserFriendlyUniqueTogether.__select__
                   & unique_together_error_for('pref_label_of', 'language_code'))

    def raise_user_exception(self):
        errors = {'': _('a preferred label in "%(lang)s" language already exists'),
                  'language_code': _('please use another language code')}
        raise ValidationError(self.entity.eid, errors, {'lang': self.entity.language_code})


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (IUserFriendlyUniqueTogether,))
    vreg.register_and_replace(IUserFriendlyUniqueTogether, adapters.IUserFriendlyUniqueTogether)



# coding: utf-8
#
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from itertools import count

from logilab.common.testlib import TestCase, require_module, unittest_main

from cubicweb import devtools # XXX ensure python path is ok
from cubes.skos import ExtEntity
from cubes.skos.rdfio import (RDFRegistryError, RDFRegistry, RDFLibRDFGraph,
                              LibRDFRDFGraph, rdf_graph_to_entities, unicode_with_language as ul)

class UnicodeWithLanguageTC(TestCase):

    def test_comparison_ul_ul(self):
        a = ul('toto', 'fr')
        b = ul('toto', 'fr')
        self.assertEqual(a, b)

        b = ul('toto', 'en')
        self.assertNotEqual(a, b)

        b = ul('titi', 'fr')
        self.assertNotEqual(a, b)

    def test_comparison_ul_and_other(self):
        a = ul('toto', 'fr')
        b = u'toto'
        self.assertNotEqual(a, b)

class RDFRegistryTC(TestCase):
    def setUp(self):
        self.xy = RDFRegistry()
        self.xy.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
        self.xy.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
        self.xy.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')

    def test_register_prefix(self):
        xy = self.xy
        self.assertEqual(xy.prefixes['dc'], 'http://purl.org/dc/elements/1.1/')
        # re-registering the same prefix is fine
        xy.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
        # though re-registering an existing prefix to a different url isn't
        self.assertRaises(RDFRegistryError,
                          xy.register_prefix, 'dc', 'http://purl.org/dc/elements/1.2/')
        # unless we explicitly tell it's ok
        xy.register_prefix('dc', 'http://purl.org/dc/elements/1.2/', overwrite=True)
        self.assertEqual(xy.prefixes['dc'], 'http://purl.org/dc/elements/1.2/')

    def test_register_etype_equivalence(self):
        xy = self.xy
        xy.register_etype_equivalence('CWUser', 'foaf:Person')
        self.assertEqual(xy.etype2rdf['CWUser'], 'http://xmlns.com/foaf/0.1/Person')
        # re-registering the same equivalence is fine
        xy.register_etype_equivalence('CWUser', 'foaf:Person')
        # though re-registering a different isn't
        self.assertRaises(RDFRegistryError,
                          xy.register_etype_equivalence, 'CWUser', 'foaf:Personne')
        # unless we explicitly tell it's ok
        xy.register_etype_equivalence('CWUser', 'foaf:Personne', overwrite=True)
        self.assertEqual(xy.etype2rdf['CWUser'], 'http://xmlns.com/foaf/0.1/Personne')
        xy.register_etype_equivalence('User', 'foaf:Person', overwrite=True)

    def test_register_attribute_equivalence(self):
        xy = self.xy
        xy.register_etype_equivalence('CWUser', 'foaf:Person')
        xy.register_attribute_equivalence('CWUser', 'login', 'dc:title')
        self.assertEqual(xy.attr2rdf[('CWUser', 'login')], 'http://purl.org/dc/elements/1.1/title')
        # re-registering the same equivalence is fine
        xy.register_attribute_equivalence('CWUser', 'login', 'dc:title')
        # though re-registering a different isn't
        self.assertRaises(RDFRegistryError,
                          xy.register_attribute_equivalence, 'CWUser', 'login', 'dc:description')
        # unless we explicitly tell it's ok
        xy.register_attribute_equivalence('CWUser', 'login', 'dc:description', overwrite=True)
        self.assertEqual(xy.attr2rdf[('CWUser', 'login')], 'http://purl.org/dc/elements/1.1/description')

    def test_register_relation_equivalence(self):
        xy = self.xy
        xy.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
        xy.register_etype_equivalence('Concept', 'skos:Concept')
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:broader')
        self.assertEqual(xy.rel2rdf[('Concept', 'in_scheme', 'ConceptScheme')],
                         set([('http://www.w3.org/2004/02/skos/core#inScheme', False)]))
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inSchema')
        self.assertEqual(xy.rel2rdf[('Concept', 'in_scheme', 'ConceptScheme')],
                        set([('http://www.w3.org/2004/02/skos/core#inScheme', False),
                             ('http://www.w3.org/2004/02/skos/core#inSchema', False)]))
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:broader')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:narrower',
                                         reverse=True)
        self.assertEqual(xy.rel2rdf[('Concept', 'broader_concept', 'Concept')],
                         set([('http://www.w3.org/2004/02/skos/core#broader', False),
                              ('http://www.w3.org/2004/02/skos/core#narrower', True)]))

    def test_predicates_for_subject_etype(self):
        xy = self.xy
        xy.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
        xy.register_etype_equivalence('Concept', 'skos:Concept')
        xy.register_attribute_equivalence('ConceptScheme', 'title', 'dc:title')
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:narrower',
                                         reverse=True)
        self.assertEqual(sorted(xy.predicates_for_subject_etype('ConceptScheme')),
                         [('title', 'http://purl.org/dc/elements/1.1/title', False)])
        self.assertEqual(sorted(xy.predicates_for_subject_etype('Concept')),
                         [('broader_concept', 'http://www.w3.org/2004/02/skos/core#narrower', True),
                          ('in_scheme', 'http://www.w3.org/2004/02/skos/core#inScheme', False)])


def dcf(string):
    return 'http://data.culture.fr/thesaurus/resource/ark:/67717/' + string


def ext_entities_to_dict(ext_entities):
    """Turn a sequence of external entities into a 2-level dict with entity
    type and extid as keys.
    """
    entities_dict = {}
    for extentity in ext_entities:
        entities_dict.setdefault(extentity.etype, {})[extentity.extid] = extentity.values
    return entities_dict


class RDFLibRDFGraphTC(TestCase):

    def test_add(self):
        graph = RDFLibRDFGraph()
        bob = graph.uri("http://example.org/people/Bob")
        knows = graph.uri("http://foaf.com/knows")
        alice = graph.uri("http://example.org/people/Alice")
        firstname = graph.uri("http://foaf.com/firstname")
        age = graph.uri("http://foaf.com/age")
        desc = graph.uri("http://dc.com/description")
        graph.add(bob, knows, alice)
        graph.add(bob, firstname, "bob")
        graph.add(bob, age, 45)
        graph.add(bob, desc, ul("man", "en"))
        graph.add(alice, firstname, "alice")
        graph.add(alice, age, 25)

        result = list(graph.objects(bob, knows))
        self.assertEqual(result, ["http://example.org/people/Alice"])
        result = list(graph.objects(bob, age))
        self.assertEqual(result, [45])
        result = list(graph.objects(bob, desc))
        self.assertEqual(result, [ul("man", "en")])
        result = list(graph.objects(alice, firstname))
        self.assertEqual(result, ["alice"])


class RDFGraphToEntitiesTC(TestCase):

    @require_module('rdflib')
    def test_rdflib(self):
        self._test(RDFLibRDFGraph())

    @require_module('RDF')
    def test_librdf(self):
        self._test(LibRDFRDFGraph())

    def _test(self, graph):
        xy = RDFRegistry()
        xy.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
        xy.register_prefix('skos', 'http://www.w3.org/2004/02/skos/core#')
        xy.register_etype_equivalence('ConceptScheme', 'skos:ConceptScheme')
        xy.register_etype_equivalence('Concept', 'skos:Concept')
        xy.register_attribute_equivalence('ConceptScheme', 'title', 'dc:title')
        xy.register_relation_equivalence('Concept', 'in_scheme', 'ConceptScheme', 'skos:inScheme')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:broader')
        xy.register_relation_equivalence('Concept', 'broader_concept', 'Concept', 'skos:narrower',
                                         reverse=True)
        xy.register_relation_equivalence('Concept', 'narrower_concept', 'Concept', 'skos:narrower')
        xy.register_relation_equivalence('Concept', 'narrower_concept', 'Concept', 'skos:broader',
                                         reverse=True)
        xy.register_relation_equivalence('Concept', 'related_concept', 'Concept', 'skos:related')
        xy.register_relation_equivalence('Label', 'pref_label_of', 'Concept', 'skos:prefLabel',
                                         reverse=True)
        xy.register_relation_equivalence('Label', 'alt_label_of', 'Concept', 'skos:altLabel',
                                         reverse=True)

        graph.load(self.datapath('siaf_matieres_shortened.xml'))
        id_count = count()

        def build_label(extentity, rtype, uris):
            for label in uris:
                yield ExtEntity('Label', next(id_count),
                                {'label': set([label]),
                                 'language_code': set([label.lang]),
                                 rtype: set([extentity.extid])})

        etypes = ('ConceptScheme', 'Concept')  #, 'Label')
        rcb = {'pref_label_of': build_label,
               'alt_label_of': build_label}
        ext_entities = rdf_graph_to_entities(xy, graph, etypes,
                                             relation_callbacks=rcb)
        e = ext_entities_to_dict(ext_entities)
        self.assertEqual(e['ConceptScheme'][dcf('Matiere')],
                         {'title': set([u"Thésaurus-matières pour l'indexation des archives locales"])})
        pref_labels = [k for k, v in e['Label'].iteritems()
                       if dcf('T1-1073') in v.get('pref_label_of', ())]
        self.assertEqual(len(pref_labels), 1)
        alt_labels = [k for k, v in e['Label'].iteritems()
                      if dcf('T1-1073') in v.get('alt_label_of', ())]
        self.assertEqual(len(alt_labels), 2)
        self.assertEqual(e['Concept'][dcf('T1-1073')],
                         {'broader_concept': set([dcf('T1-3')]),
                          'in_scheme': set([dcf('Matiere')]),
                          'related_concept': set([dcf('T1-760')]),
                         })
        self.assertEqual(e['Label'][pref_labels.pop()],
                         {'label': set([ul(u'société savante', 'fr-fr')]),
                          'language_code': set([u'fr-fr']),
                          'pref_label_of': set([dcf('T1-1073')])})


if __name__ == '__main__':
    unittest_main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Utilities for RDF import/export"""

from os.path import abspath

from cubes.skos import ExtEntity


class unicode_with_language(unicode):
    """Extend an unicode string to hold a .lang attribute as well"""

    def __new__(cls, content, lang):
        self = unicode.__new__(cls, content)
        self.lang = lang
        return self

    def __eq__(self, other):
        if isinstance(other, unicode_with_language):
            return (unicode(self), self.lang) == (unicode(other), other.lang)
        return False

    def __ne__(self, other):
        return not self == other

def normalize_uri(uri, prefixes):
    """Normalize uri by attempting to substitute prefix by the associated namespace. Returns the
    normalized URI.
    """
    try:
        prefix, value = uri.split(':', 1)
    except ValueError:
        return uri
    try:
        return prefixes[prefix] + value
    except KeyError:
        return uri

class RDFRegistryError(Exception):
    pass


class RDFRegistry(object):
    """Class inspired from yams.xy / cubes.dataio.xy and holding static information about how to
    convert from a Yams model to RDF and the other way around.
    """
    def __init__(self):
        self.prefixes = {}
        self.etype2rdf = {}
        self.attr2rdf = {}
        self.rel2rdf = {}

    def normalize_uri(self, uri):
        return normalize_uri(uri, self.prefixes)

    def register_prefix(self, prefix, namespace, overwrite=False):
        """Associate a prefix to a namespace. If the prefix is already registered to a different
        namespace, an exception is raised unless overwrite is True. Registered prefixes may be used
        in RDF snippets used in `register_*` methods.
        """
        if not overwrite and self.prefixes.get(prefix, namespace) != namespace:
            raise RDFRegistryError('prefix %r is already defined with different value %r'
                                   % (prefix, self.prefixes[prefix]))
        self.prefixes[prefix] = namespace

    def register_etype_equivalence(self, etype, rdftype, overwrite=False):
        """Associate a Yams entity type to a RDF type. If the entity type is already registered to a
        different RDF type, an exception is raised unless overwrite is True.
        """
        rdftype = self.normalize_uri(rdftype)
        if not overwrite and self.etype2rdf.get(etype, rdftype) != rdftype:
            raise RDFRegistryError('entity type %r is already associated to RDF type %r'
                                   % (etype, self.etype2rdf[etype]))
        self.etype2rdf[etype] = rdftype

    def register_attribute_equivalence(self, etype, attr, rdftype, overwrite=False):
        """Associate a Yams entity attribute to a RDF predicate. If the entity attribute is already
        registered to a different RDF type, an exception is raised unless overwrite is True.
        """
        rdftype = self.normalize_uri(rdftype)
        if not overwrite and self.attr2rdf.get((etype, attr), rdftype) != rdftype:
            raise RDFRegistryError('entity attribute %s.%r is already associated to RDF type %r'
                                   % (etype, attr, self.attr2rdf[(etype, attr)]))
        self.attr2rdf[(etype, attr)] = rdftype

    def register_relation_equivalence(self, subject_etype, rel, object_etype, rdftype,
                                      reverse=False):
        """Associate a Yams entity relation to a RDF predicated. The `reverse` flag may be used to
        indicate that in RDF, the subject and object are reversed (e.g. 'E1 yams_relation E2' is
        'E2 predicate E1 in RDF').
        """
        rdftype = self.normalize_uri(rdftype)
        self.rel2rdf.setdefault((subject_etype, rel, object_etype), set()).add((rdftype, reverse))

    def predicates_for_subject_etype(self, etype):
        """Given a yams entity type, return (yams relation type, rdf predicate uri, reverse) 3-uple
        where the entity type is subject of the yams relation. `reverse` is a boolean flag telling
        wether the relation should be expected in the opposite direction in RDF (i.e. corresponding
        entity is the object of the rdf predicate), as they are not necessarily in the same order.
        """
        for (subject_etype, attr), rdftype in self.attr2rdf.items():
            if subject_etype == etype:
                yield attr, rdftype, False
        for (subject_etype, rel, _), rdf_relations in self.rel2rdf.items():
            if subject_etype == etype:
                for (rdftype, reverse) in rdf_relations:
                    yield rel, rdftype, reverse

    def additional_object_predicates(self, etype):
        """Given a yams entity type, return (yams relation type, rdf predicate uri, reverse) 3-uple
        where the entity type is object of the yams relation. `reverse` is a boolean flag telling
        wether the relation should be expected in the opposite direction in RDF (i.e. corresponding
        entity is the object of the rdf predicate), as they are not necessarily in the same order.

        Only relations whose subject is not a registered entity types (i.e. have had a mapping
        registered by :meth:`register_etype_equivalence`) will be returned.
        """
        for (subject_etype, rel, object_etype), rdf_relations in self.rel2rdf.items():
            if object_etype == etype and subject_etype not in self.etype2rdf:
                for (rdftype, reverse) in rdf_relations:
                    yield rel, rdftype, reverse


class LibRDFRDFGraph(object):
    """redland's librdf based RDF graph"""

    def __init__(self, options_string="new='yes',hash-type='memory',dir='.'"):
        import RDF
        storage = RDF.HashStorage("test", options_string)
        self._model = RDF.Model(storage)
        self._parser = RDF.Parser('raptor')
        self._uri = RDF.Uri
        self._node = RDF.Node
        self._stmt = RDF.Statement

    def load(self, path_or_url, rdf_format=None):
        """Add RDF triplets from a file path or URL into the graph. `rdf_format` may be one of 'nt',
        'n3' or 'xml'. If not specified, it will be guessed from the file or URL extension.
        """
        if rdf_format not in (None, 'nt', 'n3', 'xml', 'rdfa', 'grddl'):
            raise ValueError('unknown rdf_format: %s' % rdf_format)
        if not ':/' in path_or_url:
            path_or_url = 'file://' + abspath(path_or_url)
        uri = self._uri(string=path_or_url)
        self._parser.parse_into_model(self._model, uri)
        self._model.sync() # in case model use a persistent storage

    def uris_for_type(self, type_uri):
        """Yield URIs of the given RDF type"""
        qs = self._stmt(subject=None,
                        predicate=self._node(uri_string='http://www.w3.org/1999/02/22-rdf-syntax-ns#type'),
                        object=self._node(uri_string=type_uri))
        for statement in self._model.find_statements(qs):
            yield str(statement.subject.uri)

    def objects(self, entity_uri, predicate_uri):
        """Yield object URIs or literals that are linked to `entity_uri` through
        `predicate_uri`.
        """
        qs = self._stmt(subject=self._node(uri_string=entity_uri),
                        predicate=self._node(uri_string=predicate_uri),
                        object=None)
        for statement in self._model.find_statements(qs):
            if statement.object.is_literal():
                string_value, language, datatype_uri = statement.object.literal
                if language is not None:
                    yield unicode_with_language(string_value, language)
                else:
                    yield string_value #XXX type according to datatype_uri
            else:
                yield str(statement.object.uri)

    def subjects(self, predicate_uri, entity_uri):
        """Yield subject URIs that are linked to `entity_uri` through
        `predicate_uri`.
        """
        qs = self._stmt(subject=None,
                        predicate=self._node(uri_string=predicate_uri),
                        object=self._node(uri_string=entity_uri))
        for statement in self._model.find_statements(qs):
            yield str(statement.subject.uri)


class RDFLibRDFGraph(object):
    """rdflib based RDF graph (http://rdflib.readthedocs.org)"""

    def __init__(self):
        import rdflib
        self.uri = rdflib.URIRef
        self._namespace = rdflib.namespace
        self._literal = rdflib.Literal
        self._guess_format = rdflib.util.guess_format
        self._graph = rdflib.ConjunctiveGraph()
        from rdflib.plugin import register, Parser
        register('text/rdf+n3', Parser, 'rdflib.plugins.parsers.notation3', 'N3Parser')

    def add(self, subj, predicate, obj):
        """Add statement to graph. subject and predicate are expected to be URIs and object may
        be either a URI or a literal value.
        """
        assert isinstance(subj, self.uri)
        assert isinstance(predicate, self.uri)
        if not isinstance(obj, self.uri):
            if isinstance(obj, unicode_with_language):
                obj = self._literal(obj, lang=obj.lang)
            else:
                obj = self._literal(obj)
        self._graph.add((subj, predicate, obj))


    def load(self, path_or_url, rdf_format=None):
        """Add RDF triplets from a file stream, path or URL into the graph. `rdf_format` may be one
        of 'nt', 'n3' or 'xml'. If not specified, it will be guessed from the file or URL extension.
        """
        assert rdf_format in (None, 'nt', 'n3', 'xml', 'rdfa', 'grddl'), rdf_format
        if rdf_format is None and isinstance(path_or_url, basestring):
            rdf_format = self._guess_format(path_or_url)
        self._graph.parse(path_or_url, format=rdf_format)

    def uris_for_type(self, type_uri):
        """Return an iterator on URIs of the given RDF type"""
        for subj in self._graph.subjects(self._namespace.RDF.type, self.uri(type_uri)):
            yield unicode(subj)

    def objects(self, entity_uri, predicate_uri):
        """Return an iterator on object URIs or literals that are linked to `entity_uri` through
        `predicate_uri`.
        """
        for obj in self._graph.objects(self.uri(entity_uri), self.uri(predicate_uri)):
            if isinstance(obj, self.uri):
                yield unicode(obj)
            elif obj.language is not None:
                yield unicode_with_language(obj.toPython(), obj.language)
            else:
                yield obj.toPython()

    def subjects(self, predicate_uri, entity_uri):
        """Return an iterator on subject URIs that are linked to `entity_uri` through
        `predicate_uri`.
        """
        for subj in self._graph.subjects(self.uri(predicate_uri), self.uri(entity_uri)):
            yield unicode(subj)


def rdf_graph_to_entities(reg, graph, etypes, relation_callbacks=None):
    """Turns RDF data into an transitional ExtEntity representation that may
    be then imported into some database easily.

    Mapping is done using a RDF registry `reg` (see :class:`RDFRegistry`).

    A `relation_callbacks` dictionary may be given, containing some relation type as key and a
    callback function to be called in place of the default behaviour. This function will be given
    (extentity, relation type, uri or literal values) as arguments.
    """
    def default_callback(extentity, rtype, uris):
        uris = set(uris)
        if uris:
            extentity.values[rtype] = set(uris)
        return ()

    if relation_callbacks is None:
        relation_callbacks = {}
    for etype in etypes:
        type_uri = reg.etype2rdf[etype]
        for uri in graph.uris_for_type(type_uri):
            extentity = ExtEntity(etype, uri)
            for rtype, predicate_uri, reverse in reg.predicates_for_subject_etype(etype):
                if reverse:
                    uris = graph.subjects(predicate_uri, uri)
                else:
                    uris = graph.objects(uri, predicate_uri)
                callback = relation_callbacks.get(rtype, default_callback)
                for edict in callback(extentity, rtype, uris):
                    yield edict
            for rtype, predicate_uri, reverse in reg.additional_object_predicates(etype):
                if reverse:
                    uris = graph.objects(uri, predicate_uri)
                else:
                    uris = graph.subjects(predicate_uri, uri)
                callback = relation_callbacks.get(rtype, default_callback)
                for edict in callback(extentity, rtype, uris):
                    yield edict
            yield extentity
